/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

/*
 *
 * Authors: Paolo Andreetto, <paolo.andreetto@pd.infn.it>
 *
 */

package org.glite.ce.commonj.authn;

import java.io.File;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.Attribute;
import javax.naming.directory.BasicAttributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchResult;
import javax.security.auth.Subject;
import javax.servlet.http.HttpServletRequest;
import javax.xml.namespace.QName;

import org.apache.axis.AxisFault;
import org.apache.axis.MessageContext;
import org.apache.axis.description.OperationDesc;
import org.apache.axis.handlers.BasicHandler;
import org.apache.axis.transport.http.HTTPConstants;
import org.apache.log4j.Logger;

import org.glite.ce.commonj.authz.AuthorizationException;
import org.glite.ce.commonj.authz.AuthorizationHandler;
import org.glite.ce.commonj.utils.CEUtils;
import org.glite.ce.faults.AuthenticationFault;
import org.glite.voms.VOMSAttribute;

public class AuthenticationStorageHandler extends BasicHandler {
    private static final Logger logger = Logger.getLogger(AuthenticationStorageHandler.class.getName());
    private static final long serialVersionUID = 1202907233;
    private static String userProxyPath = null;

    public AuthenticationStorageHandler() {
        super();
    }

    public void init() {
        super.init();

        if (userProxyPath == null) {
            DirContext rootCtx = null;
            Hashtable<String, String> env = new Hashtable<String, String>(0);
            env.put(Context.INITIAL_CONTEXT_FACTORY, "org.apache.naming.java.javaURLContextFactory");
            env.put(Context.URL_PKG_PREFIXES, "org.apache.naming");

            try {
                Context context = new InitialContext(env);
                context = (Context) context.lookup("java:comp/env");

                String confProvider = (String) context.lookup("configuration_provider");
                String confProviderURL = (String) context.lookup("configuration_provider_url");

                if (confProvider == null) {
                    logger.error("Missing configuration_provider parameter");
                    return;
                }

                if (confProviderURL == null) {
                    logger.error("Missing configuration_provider_url parameter");
                    return;
                }

                env.put(Context.INITIAL_CONTEXT_FACTORY, confProvider);
                env.put(Context.PROVIDER_URL, confProviderURL);
                rootCtx = (DirContext) new InitialDirContext(env);

                Attribute attribute = rootCtx.getAttributes("").get("cream_scratch_dir");
                if (attribute == null) {
                    throw new Exception("cream_scratch_dir attribute not found!");
                }

                userProxyPath = (String) attribute.get(0) + File.separator + "user_proxy" + File.separator;
                logger.debug("userProxyPath = " + userProxyPath);

            } catch (Exception nEx) {
                logger.error(nEx.getMessage(), nEx);
                throw new RuntimeException("Cannot configure AuthorizationHandler");
            }
        }
    }
    
    public void invoke(MessageContext msgContext) throws AxisFault {
        if(msgContext == null) {
            return;
        }
        
        String dn = (String) msgContext.getProperty(AuthorizationHandler.USERDN_RFC2253_LABEL);
        if (dn == null) {
            throw getAuthenticationFault("User DN not found!");
        }        

        String userVO = CEUtils.getUserDefaultVO();
        if (userVO == null ) {
            logger.error("Default VO not defined in msgContext");
            throw getAuthenticationFault("Default VO not defined in msgContext");
        }
        
        List vomsList = (List) msgContext.getProperty(AuthorizationHandler.USER_VOMSATTRS_LABEL);
        if(vomsList == null) {
            logger.error("USER_VOMSATTRS_LABEL not defined in msgContext");
            throw getAuthenticationFault("USER_VOMSATTRS_LABEL not defined in msgContext");
        }

        Iterator item = vomsList.iterator();
        VOMSAttribute vomsAttr = null;
        String fqan = null;
 
        if(item != null) {
            while (item.hasNext()) {
                vomsAttr = (VOMSAttribute) item.next();
 
                if (userVO.equals(vomsAttr.getVO())) {
                    List fqanList = vomsAttr.getListOfFQAN();
                    if(fqanList.size() > 0) {
                        fqan = fqanList.iterator().next().toString();
                        fqan = normalize(fqan);
                    }
                
                    break;
                }
            }
        }
        
        String authNProxyPath = userProxyPath + normalize(dn + fqan);
        
        try {
            CEUtils.makeDir(authNProxyPath);
        } catch (Exception e) {
            logger.error(e);
            throw getAuthenticationFault("cannot create the authN proxy diri [" + authNProxyPath + "]!");
        }
 
        String fullPath = authNProxyPath + File.separator + "authN.proxy";

        File authProxyFile = new File(fullPath);
        if (!authProxyFile.exists() || (Calendar.getInstance().getTimeInMillis() - authProxyFile.lastModified() > 600000)) {
            String authNproxy = null;

            try {
                authNproxy = CEUtils.getPEM((X509Certificate[])msgContext.getProperty(AuthorizationHandler.USER_CERTCHAIN_LABEL));
            } catch(Exception e) {
                logger.error(e);
                throw getAuthenticationFault("cannot retrieve the authN proxy!");
            }

            if(authNproxy == null) {
                throw getAuthenticationFault("the authN proxy is empty!");
            }

            try {
                 CEUtils.makeFile(fullPath, authNproxy, false, true);

                 logger.debug("Written AuthNProxy: " + fullPath + "  " + Calendar.getInstance().getTime());
            } catch (Exception e) {
                logger.error(e);
                throw getAuthenticationFault("cannot write the authN proxy to file: " + e.getMessage());
            }
        }        
    }

    protected AxisFault getAuthenticationFault(String message) {
        AuthenticationFault fault = new AuthenticationFault();
        fault.setMethodName("invoke");
        fault.setDescription(message);
        fault.setErrorCode("0");
        fault.setFaultCause(message);
        fault.setTimestamp(Calendar.getInstance());
        return fault;
    }

    protected String normalize(String s) {
        if (s != null) {
            return s.replace('=', '_').replace(' ', '_').replace('/', '_').replace(',', '_').replace('@', '_');
        }
        return null;
    }
}
