/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

/*
 * 
 * Authors: Silvano Squizzato (silvano.squizzato@pd.infn.it)
 *
 */

package org.glite.ce.commonj.authn.db;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Calendar;

import org.apache.log4j.Logger;
import org.glite.ce.common.db.DatabaseException;
import org.glite.ce.common.db.DatasourceManager;
import org.glite.ce.commonj.authn.AuthNException;
import org.glite.ce.commonj.authn.AuthNInterface;

/**
 * Class managing a command queue in the database.
 * 
 */
public class AuthenticationDBManager implements AuthNInterface {
    /** The logger */
    private final static Logger logger = Logger.getLogger(AuthenticationDBManager.class);

    /** The datasource name */
    private String datasourceName = "java:comp/env/jdbc/authndb";;

    /**
     * DAO for managing Authentication table.
     */
    private AuthenticationTable authenticationTable = null;

    
    /**
     * Constructor.
     * @param datasourceName The datasource name.
     * @throws CommandQueueException
     */
    public AuthenticationDBManager(String datasourceName) throws AuthNException {
        init(datasourceName);
    }
    
    /**
    /**
     * Constructor.
     * @throws CommandQueueException
     */
    public AuthenticationDBManager() throws AuthNException {
    }
    
    /**
     * Initialises the datasource.
     * @param datasourceName The datasource.
     * @throws AuthNException
     */
    private void init(String datasourceName) throws AuthNException {
        // Init datasource name
        this.datasourceName = datasourceName;
        // Init DAOs
        authenticationTable = new AuthenticationTable();

    }
    
    /**
     * Returns the connection to the database.
     * @return The connection to the database.
     * @throws DatabaseException 
     */
    private Connection getConnection() throws DatabaseException {
       return DatasourceManager.getConnection(datasourceName);
    }

    /**
     * @see org.glite.ce.commonj.authn.AuthNInterface#deleteUserAuthNCertificate(java.lang.String)
     */
    public void deleteUserAuthNCertificate(String userId)
            throws AuthNException, IllegalArgumentException {
        Connection connection = null;
        logger.debug("deleteCommand begin");
        
        try {
            connection = getConnection();
        } catch (DatabaseException e) {
            logger.error("Problem in getting connection: " + e.getMessage());
            throw new AuthNException(e.getMessage());
        }
            
        try {    
            if (connection != null) {   
                // Delete
                authenticationTable.executeDelete(userId, connection);
                
                // Commit
                connection.commit();
            } else {
                logger.error("Connection is null"); 
                throw new AuthNException("Connection is null");
            }
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                connection.rollback();
                logger.error("Rollback performed");
                throw new AuthNException(sqle.getMessage());
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new AuthNException(sqle1.getMessage());
            }
        } finally {
            if (connection != null) {
                try { 
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new AuthNException(sqle2.getMessage());
                }
            }
        }
        logger.debug("deleteCommand end");
    }

    /**
     * Retrieves the authentication in for a given <code>userId</code>.
     * @param userId The <code>userId</code>.
     * @return
     * @throws AuthNException
     * @throws IllegalArgumentException
     */
    public AuthenticationInfo getAuthenticationInfo(String userId) throws AuthNException,
            IllegalArgumentException {
        Connection connection = null;
        AuthenticationInfo authenticationInfo = null;
        logger.debug("getUserAuthNCertificate begin");
        
        try {
            connection = getConnection();
        } catch (DatabaseException e) {
            logger.error("Problem in getting connection: " + e.getMessage());
            throw new AuthNException(e.getMessage());
        }
            
        try {
            
            if (connection != null) {
                // Select
                authenticationInfo = authenticationTable.executeSelect(userId, connection);
                
                // Commit
                connection.commit();
            } else {
                logger.error("Connection is null");
                throw new AuthNException("Connection is null");
            }
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                connection.rollback();
                logger.error("Rollback performed");
                throw new AuthNException(sqle.getMessage());
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new AuthNException(sqle1.getMessage());
            }
        } finally {
            if (connection != null) {
                try { 
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new AuthNException(sqle2.getMessage());
                }
            }
        }
        logger.debug("getUserAuthNCertificate end");
        return authenticationInfo;
    }

    /**
     * @see org.glite.ce.commonj.authn.AuthNInterface#getUserAuthNCertificateExpirationTime(java.lang.String)
     */
    public Calendar getUserAuthNCertificateExpirationTime(String userId)
            throws AuthNException, IllegalArgumentException {
        Calendar expirationTime = null;
        AuthenticationInfo info = getAuthenticationInfo(userId);
        if (info != null) {
            expirationTime = info.getExpirationTime();
        }
        return expirationTime;
    }
    
    /**
     * @see org.glite.ce.commonj.authn.AuthNInterface#getUserAuthNCertificate(java.lang.String)
     */
    public String getUserAuthNCertificate(String userId) throws AuthNException, IllegalArgumentException {
        String certificate = null;
        AuthenticationInfo info = getAuthenticationInfo(userId);
        if (info != null) {
            certificate = info.getAuthNCert();
        }
        return certificate;
    }
    
    /**
     * @see org.glite.ce.commonj.authn.AuthNInterface#getLastUserAuthNCertificateUpdateTime(java.lang.String)
     */
    public Calendar getLastUserAuthNCertificateUpdateTime(String userId)
            throws AuthNException, IllegalArgumentException {
        Calendar modificationTime = null;
        AuthenticationInfo info = getAuthenticationInfo(userId);
        if (info != null) {
            modificationTime = info.getModificationTime();
        }
        return modificationTime;
    }
    
    /**
     * @see org.glite.ce.commonj.authn.AuthNInterface#isUserAuthNCertificateValid(java.lang.String)
     */
    public boolean isUserAuthNCertificateValid(String userId)
            throws AuthNException, IllegalArgumentException {
        // TODO Auto-generated method stub
        return true;
    }

    /**
     * @see org.glite.ce.commonj.authn.AuthNInterface#setUserAuthNCertificate(java.lang.String, java.lang.String, java.util.Calendar)
     */
    public void setUserAuthNCertificate(String userId, String authNCert,
            Calendar expirationTime) throws AuthNException,
            IllegalArgumentException {
        Connection connection = null;
        logger.debug("setUserAuthNCertificate begin");
        
        try {
            connection = getConnection();
        } catch (DatabaseException e) {
            throw new AuthNException(e.getMessage());
        }
            
        try {    
            if (connection != null) {
                if (isUserIdPresent(userId)) {
                    // Update
                    authenticationTable.executeUpdate(userId, authNCert, expirationTime, connection); 
                } else {
                    // Insert
                    authenticationTable.executeInsert(userId, authNCert, expirationTime, connection); 
                }

                // Commit
                connection.commit();
            } else {
                logger.error("Connection is null");
                throw new AuthNException("Connection is null");
            }
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                connection.rollback();
                logger.error("Rollback performed");
                throw new AuthNException(sqle.getMessage());
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new AuthNException(sqle1.getMessage());
            }
        } finally {
            if (connection != null) {
                try { 
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new AuthNException(sqle2.getMessage());
                }
            }
        }
        logger.debug("setUserAuthNCertificate end");        
    }
    
    /**
     * Checks if a <code>userId</code> already has a certificate registered.
     * userId The userId.
     * @return Returns true if the <code>userId</code> exists, false otherwise.
     * @throws AuthNException
     */
    public boolean isUserIdPresent(String userId) throws AuthNException {
        Connection connection = null;
        boolean result = false;
        logger.debug("isQueueNamePresent begin");
        
        try {
            connection = getConnection();
        } catch (DatabaseException de) {
            logger.error("Problem in getting connection: " + de.getMessage());
            throw new AuthNException(de.getMessage());
        }
        
        try {
            if (connection != null) {
                // Searches for a given userId
                result = authenticationTable.executeIsUserIdPresent(userId, connection);
                logger.debug("Executed executeIsUserIdPresent result=" + result); 

                // Commit
                connection.commit();
            } else {
                logger.error("Connection is null");
                throw new AuthNException("Connection is null");
            }
        } catch (SQLException sqle) {
            try {
                logger.error("Operation failed: " + sqle.getMessage());
                if (connection != null) {
                    connection.rollback();
                    logger.error("Rollback performed");
                }
            } catch (SQLException sqle1) {
                logger.error("Rollback failed: " + sqle1.getMessage());
                throw new AuthNException(sqle1.getMessage());
            }
            throw new AuthNException(sqle.getMessage());
        } finally {
            if (connection != null) {
                try {
                    connection.close();
                } catch (SQLException sqle2) {
                    logger.error("Problem in closing connection: " + sqle2.getMessage());
                    throw new AuthNException(sqle2.getMessage());
                }
            }
        }
        logger.debug("isQueueNamePresent end");
        return result;
    }
}
