/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

/*
 * 
 * Authors: Silvano Squizzato (silvano.squizzato@pd.infn.it)
 *
 */

package org.glite.ce.commonj.authn.db;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.TimeZone;

import org.apache.log4j.Logger;
import org.glite.ce.common.db.AbstractTable;
import org.glite.ce.common.db.Query;



/**
 * Class representing the Data Access Object (DAO) to the authentication table in the database.
 *
 */
public class AuthenticationTable extends AbstractTable implements AuthenticationTableInterface  {
    
    /** The logger */
    final static Logger logger = Logger.getLogger(AuthenticationTable.class);

    /** The table name */
    public final static String NAME_TABLE = "authentication";

    // The table columns
    public final static String USER_ID_FIELD = "userId";
    public final static String AUTHENTICATION_CERTIFICATE_FIELD = "authNCert";
    public final static String EXPIRATION_TIME_FIELD = "expirationTime";
    public final static String MODIFICATION_TIME_FIELD = "modificationTime";
    
    /** Query name to check if a <code>userId</code> exists in the db */
    public static final String IS_USERID_PRESENT = "IS_USERID_PRESENT";
     
    /** Query to check if a <code>userId</code> exists in the db */
    protected Query isUserIdPresentQuery = null;
    
    /**
     * @see org.glite.ce.commonj.authn.db.AuthenticationTableInterface#executeDelete(java.lang.String, java.sql.Connection)
     */
    public void executeDelete(String commandId, Connection connection)
            throws SQLException, IllegalArgumentException {
        logger.debug("executeDelete begin");
        PreparedStatement pstmt = null;
        try {
            pstmt = getPreparedStatement(getDeleteQuery(), connection);
            if (pstmt != null) {
                if (checkBeforeDelete(commandId)) {
                    fillDeletePreparedStatement(commandId, pstmt);
                    int rowCount = pstmt.executeUpdate();
                    logger.debug("deleted " + rowCount + " rows");
                }
            } else {
                logger.error("prepared statement null");
            }
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
        }
        logger.debug("executeDelete end");
    }

    
    /**
     * @see org.glite.ce.commonj.authn.db.AuthenticationTableInterface#executeInsert(java.lang.String, java.lang.String, java.util.Calendar, java.sql.Connection)
     */
    public void executeInsert(String userId, String authNCert,
            Calendar expirationTime, Connection connection)
            throws SQLException, IllegalArgumentException {
        logger.debug("executeInsert begin");
        PreparedStatement pstmt = null;
        try {
            pstmt = getPreparedStatement(getInsertQuery(), connection);
            if (pstmt != null) {
                if (checkBeforeInsert(userId, authNCert, expirationTime)) {
                    fillInsertPreparedStatement(userId, authNCert, expirationTime, pstmt);
                    int rowCount = pstmt.executeUpdate();
                    logger.debug("inserted " + rowCount + " rows");
                }
            } else {
                logger.error("prepared statement null");
            }
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
        }
        logger.debug("executeInsert end");        
    }

    /**
     * Returns true if a <code>userId</code> already exists.
     * @param userId The <code>userId</code>.
     * @param connection The connection to db.
     * @return
     * @throws SQLException
     * @throws IllegalArgumentException
     */
    public boolean executeIsUserIdPresent(String userId, Connection connection) throws SQLException, IllegalArgumentException {
        boolean result = false;
        logger.debug("executeIsUserIdPresent begin");
        PreparedStatement pstmt = null;
        ResultSet rset = null;
        try {
            // Check if a queue name is registered
            pstmt = getPreparedStatement(getIsUserIdPresentQuery(), connection);
            if (pstmt != null) {
                fillIsUserIdPresentPreparedStatement(userId, pstmt);
                rset = pstmt.executeQuery();
                if (rset != null) {
                    rset.next();
                    if (rset.getRow() > 0) {
                        long count = rset.getLong(1);
                        if (count > 0) {
                            result = true; 
                        }
                    } 
                }
                logger.debug("check result = " + result);
            } else {
                logger.error("prepared statement null");
            }
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }
        logger.debug("executeIsUserIdPresent end");
        return result;
    }
    
    /**
     * @see org.glite.ce.commonj.authn.db.AuthenticationTableInterface#executeSelect(java.lang.String, java.sql.Connection)
     */
    public AuthenticationInfo executeSelect(String userId, Connection connection)
            throws SQLException, IllegalArgumentException {
        logger.debug("executeSelect begin");
        AuthenticationInfo authenticationInfo = null;
        PreparedStatement pstmt = null;
        ResultSet rset = null;
        try {
            pstmt = getPreparedStatement(getSelectQuery(), connection);
            if (pstmt != null) {
                if (checkBeforeSelect(userId)) {
                    fillSelectPreparedStatement(userId, pstmt);
                    rset = pstmt.executeQuery();
                    if (rset != null) {
                        if (rset.next()) {
                            authenticationInfo = new AuthenticationInfo();
                            authenticationInfo.setUserId(userId);
                            authenticationInfo.setAuthNCert(rset.getString(1));
        
                            
                            Timestamp timestamp = rset.getTimestamp(2);
                            Calendar calendar = null;
                            if (timestamp != null) {
                                calendar = Calendar.getInstance();
                                calendar.setTimeInMillis(timestamp.getTime());
                            } 
                            authenticationInfo.setExpirationTime(calendar);
                            
                            timestamp = rset.getTimestamp(3);
                            calendar = null;
                            if (timestamp != null) {
                                calendar = Calendar.getInstance();
                                calendar.setTimeInMillis(timestamp.getTime());
                            } 
                            authenticationInfo.setModificationTime(calendar);
                        }
                    }
                } else {
                    logger.error("prepared statement null");
                }
            }
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
            if (rset != null) {
                try {
                    rset.close();
                } catch (SQLException sqle2) {
                    logger.error(sqle2.getMessage());
                }
            }
        }
        logger.debug("executeSelect end");
        return authenticationInfo;
    }
    
    /**
     * @see org.glite.ce.commonj.authn.db.AuthenticationTableInterface#executeUpdate(java.lang.String, java.lang.String, java.util.Calendar, java.sql.Connection)
     */
    public void executeUpdate(String userId, String authNCert,
            Calendar expirationTime, Connection connection)
            throws SQLException, IllegalArgumentException {
        AuthenticationInfo authenticationInfo = new AuthenticationInfo(userId, authNCert, expirationTime, null);
        executeUpdate(authenticationInfo, connection);
    }
    
    /**
     * Updates the authentication information.
     * @param authenticationInfo The authentication information.
     * @param connection The connection to db.
     * @throws SQLException
     * @throws IllegalArgumentException
     */
    public void executeUpdate(AuthenticationInfo authenticationInfo, Connection connection)
            throws SQLException, IllegalArgumentException {
        logger.debug("executeUpdate begin");
        PreparedStatement pstmt = null;
        try {
            pstmt = getPreparedStatement(getUpdateQuery(authenticationInfo), connection);
            if (pstmt != null) {
                if (checkBeforeUpdate(authenticationInfo)) {
                    Calendar now = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
                    authenticationInfo.setModificationTime(now);
                    fillUpdatePreparedStatement(authenticationInfo, pstmt);
                    int rowCount = pstmt.executeUpdate();
                    logger.debug("updated " + rowCount + " rows");
                }
            } else {
                logger.error("prepared statement null");
            }
        } catch (SQLException sqle) {
            throw sqle;
        } finally {
            if (pstmt != null) {
                try {
                    pstmt.close();
                } catch (SQLException sqle1) {
                    logger.error(sqle1.getMessage());
                }
            }
        }
        logger.debug("executeUpdate end");
        
    }
    
    /**
     * Returns the query for inserting
     * @return The query for inserting.
     */
    protected Query getInsertQuery() {
        if (insertQuery == null) {
            StringBuffer query = new StringBuffer();
            query.append("insert into ");
            query.append(NAME_TABLE);
            query.append(" ( ");
            query.append(USER_ID_FIELD + ", ");
            query.append(AUTHENTICATION_CERTIFICATE_FIELD  + ", ");
            query.append(EXPIRATION_TIME_FIELD  + ", ");
            query.append(MODIFICATION_TIME_FIELD);
            query.append(" ) ");
            query.append("values(?, ?, ?, ?)");
            insertQuery = new Query();
            insertQuery.setName(Query.INSERT_STATEMENT);
            insertQuery.setStatement(query.toString());
        }
        return insertQuery;
    }
    
    /**
     * Returns the query for deleting.
     * @return The query for deleting.
     */
    protected Query getDeleteQuery() {
        if (deleteQuery == null) {
            StringBuffer query = new StringBuffer();
            query.append("delete from ");
            query.append(NAME_TABLE);
            query.append(" where ");
            query.append(USER_ID_FIELD);
            query.append(" = ");
            query.append(" ? ");
            deleteQuery = new Query();
            deleteQuery.setName(Query.DELETE_STATEMENT);
            deleteQuery.setStatement(query.toString());
        }
        return deleteQuery;
    }
    
    /**
     * Returns the query string for updating.
     * @param authenticationInfo The authentication info to be updated.
     * @return The query string for updating.
     */
    protected Query getUpdateQuery(AuthenticationInfo authenticationInfo) {
        if (updateQuery == null) {
            StringBuffer query = new StringBuffer();
            query.append("update ");
            query.append(NAME_TABLE);
            query.append(" set ");
            query.append(AUTHENTICATION_CERTIFICATE_FIELD);
            query.append(" = ");
            query.append(" ? ");
            query.append(" , ");
            query.append(EXPIRATION_TIME_FIELD);
            query.append(" = ");
            query.append(" ? ");
            query.append(" , ");
            query.append(MODIFICATION_TIME_FIELD);
            query.append(" = ");
            query.append(" ? ");
            query.append(" where ");
            query.append(USER_ID_FIELD);
            query.append(" = ");
            query.append(" ? ");
            updateQuery = new Query();
            updateQuery.setName(Query.UPDATE_STATEMENT);
            updateQuery.setStatement(query.toString());
        }
        return updateQuery;
    }
    
    /**
     * Returns the query for selecting.
     * @return The query for selecting.
     */
    protected Query getSelectQuery() {
        if (selectQuery == null) {
            StringBuffer query = new StringBuffer();
            query.append("select ");
            query.append(AUTHENTICATION_CERTIFICATE_FIELD + ", ");
            query.append(EXPIRATION_TIME_FIELD  + ", ");
            query.append(MODIFICATION_TIME_FIELD );
            query.append(" from ");
            query.append(NAME_TABLE);
            query.append(" where ");
            query.append(USER_ID_FIELD);
            query.append(" = ");
            query.append(" ? ");
            selectQuery = new Query();
            selectQuery.setName(Query.SELECT_STATEMENT);
            selectQuery.setStatement(query.toString());
        }
        return selectQuery;
    }
    
    /**
     * Returns the query for checking if a <code>userId</code> is registered.
     * @return The query for for checking if a <code>userId</code> is registered.
     */
    protected Query getIsUserIdPresentQuery() {
        if (isUserIdPresentQuery == null) {
            StringBuffer query = new StringBuffer();
            query.append("select");
            query.append(" ");
            query.append("count(*) ");
            query.append(" from ");
            query.append(NAME_TABLE);
            query.append(" where ");
            query.append(USER_ID_FIELD);
            query.append(" = ");
            query.append(" ? ");
            isUserIdPresentQuery = new Query();
            isUserIdPresentQuery.setName(IS_USERID_PRESENT);
            isUserIdPresentQuery.setStatement(query.toString());
        }
        return isUserIdPresentQuery;
    }  
    
 
    /**
     * Fills the prepared statement with proper values.
     * @param userId The user identifier whose certificate is to be inserted.
     * @param authNCert The user authentication certificate.
     * @param expirationTime The expiration time.
     * @param pstmt The statement to be executed.
     * @throws SQLException 
     */
    public void fillInsertPreparedStatement(String userId, String authNCert, Calendar expirationTime, PreparedStatement pstmt) throws SQLException {
        if (pstmt != null) {
           pstmt.setString(1, userId);
           pstmt.setString(2, authNCert);
           pstmt.setTimestamp(3, new Timestamp(expirationTime.getTimeInMillis()));
           pstmt.setTimestamp(4, new Timestamp((new GregorianCalendar(TimeZone.getTimeZone("GMT"))).getTimeInMillis()));
           logger.debug("insert statetement = " + pstmt.toString());
        }
    }
    
    /**
     * Fills the prepared statement with proper values.
     * @param userId The <code>userId</code> to be used.
     * @param pstmt The statement to be executed.
     * @throws SQLException 
     */
    public void fillDeletePreparedStatement(String userId, PreparedStatement pstmt) throws SQLException  {
        if (pstmt != null) {
            pstmt.setString(1, userId);
            logger.debug("delete statetement = " + pstmt.toString());
        }
    }
    
    /**
     * Fills the prepared statement with proper values.
     * @param cmd The authentication info to be used.
     * @param pstmt The statement to be executed.
     * @throws SQLException 
     */
    public void fillUpdatePreparedStatement(AuthenticationInfo authenticationInfo, PreparedStatement pstmt) throws SQLException {
        if (pstmt != null && authenticationInfo != null) {
            pstmt.setString(1, authenticationInfo.getAuthNCert());
            
            Timestamp expirationTime = null;
            if (authenticationInfo.getExpirationTime() != null) {
                expirationTime = new Timestamp(authenticationInfo.getExpirationTime().getTimeInMillis());
            }
            pstmt.setTimestamp(2, expirationTime);
            
            Timestamp modificationTime = null;
            if (authenticationInfo.getModificationTime() != null) {
                modificationTime = new Timestamp(authenticationInfo.getModificationTime().getTimeInMillis());
            }
            pstmt.setTimestamp(3, modificationTime);
            pstmt.setString(4, authenticationInfo.getUserId());

            logger.debug("update statetement = " + pstmt.toString());
        }
    }
    
    /**
     * Fills the prepared statement with proper values.
     * @param userId The <code>userId</code> to be used.
     * @param pstmt The statement to be executed.
     * @throws SQLException 
     */
    public void fillSelectPreparedStatement(String userId, PreparedStatement pstmt) throws SQLException  {
        if (pstmt != null) {
            pstmt.setString(1, userId);
            logger.debug("select statetement = " + pstmt.toString());
          }
    }
    
    /**
     * Fills the prepared statement with proper values.
     * @param userId The <code>userId</code>.
     * @param pstmt The statement to be executed.
     * @throws SQLException 
     */
    public void fillIsUserIdPresentPreparedStatement(String userId, PreparedStatement pstmt) throws SQLException {
        if (pstmt != null && userId != null) {
            pstmt.setString(1, userId);
            logger.debug("isUserIdPresent statement = " + pstmt.toString());
        }
    }
    
    /**
     * Returns true if the check is OK.
     * @param userId The <code>userId</code> to be checked.
     * @return true if the check is OK
     * @throws IllegalArgumentException if mandatory conditions are violated.
     */
    protected boolean checkBeforeDelete(String userId) throws IllegalArgumentException{
        boolean result = true;
        if (userId == null) {
            throw new IllegalArgumentException("The userId is null");
        } else {
            if (userId.length() == 0) {
                throw new IllegalArgumentException("The userId = " + userId + " is invalid");
            }
        }
        return result;
    }
    
    /**
     * Returns true if the check is OK.
     * @param userId The user identifier whose certificate is to be inserted.
     * @param authNCert The user authentication certificate.
     * @return true if the check is OK.
     * @throws IllegalArgumentException if mandatory conditions are violated.
     */
    protected boolean checkBeforeInsert(String userId, String authNCert, Calendar expirationTime) throws IllegalArgumentException{
        boolean result = true;
        if (userId == null) {
            throw new IllegalArgumentException("The userId is null");
        }
        if (authNCert == null) {
            throw new IllegalArgumentException("The authentication certificate is null");
        }
        if (expirationTime == null) {
            throw new IllegalArgumentException("The expiration time is null");
        }
        return result;
    }
    
    /**
     * Returns true if the check is OK.
     * @param authenticationInfo The authentication info to be checked.
     * @return true if the check is OK
     * @throws IllegalArgumentException if mandatory conditions are violated.
     */
    protected boolean checkBeforeUpdate(AuthenticationInfo authenticationInfo) throws IllegalArgumentException{
        boolean result = true;
        if (authenticationInfo == null) {
            throw new IllegalArgumentException("The authenticationInfo object is null");
        } else {
            if (authenticationInfo.getUserId() == null || authenticationInfo.getUserId().length() == 0) {
                throw new IllegalArgumentException("The userId = " + authenticationInfo.getUserId() + " is invalid");
            }
            if (authenticationInfo.getAuthNCert() == null || authenticationInfo.getAuthNCert().length() == 0) {
                throw new IllegalArgumentException("The certificate = " + authenticationInfo.getAuthNCert() + " is invalid");
            }
            if (authenticationInfo.getExpirationTime() == null) {
                throw new IllegalArgumentException("The expiration time = " + authenticationInfo.getExpirationTime() + " is invalid");
            }
        }
        return result;
    }
    
    /**
     * Returns true if the check is OK.
     * @param userId The <code>userId</code> to be checked.
     * @return true if the check is OK
     * @throws IllegalArgumentException if mandatory conditions are violated.
     */
    protected boolean checkBeforeSelect(String userId) throws IllegalArgumentException{
        boolean result = true;
        if (userId == null) {
            throw new IllegalArgumentException("The userId is null");
        } else {
            if (userId.length() == 0) {
                throw new IllegalArgumentException("The userId = " + userId + " is invalid");
            }
        }
        return result;
    }
 }
