/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

/*
 * 
 * Authors: Paolo Andreetto, <paolo.andreetto@pd.infn.it>
 *
 */

package org.glite.ce.commonj.authz;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Calendar;
import java.util.HashSet;
import java.util.Hashtable;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.apache.axis.AxisFault;
import org.apache.axis.MessageContext;
import org.apache.axis.handlers.BasicHandler;
import org.apache.log4j.Logger;
import org.glite.ce.faults.AuthorizationFault;
import org.glite.security.SecurityInfo;
import org.glite.security.SecurityInfoContainer;
import org.glite.security.util.axis.InitSecurityContext;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

public class AdminHandler
    extends BasicHandler {

    private class AdminConfigParser
        extends DefaultHandler {

        public void startElement(String uri, String name, String qName, Attributes attributes)
            throws SAXParseException {
            if (qName.equals("adminlist")) {
                String fAttr = attributes.getValue("filename");
                if (fAttr == null) {
                    throw new SAXParseException("Missing filename for adminlist", null);
                }

                adminFile = new File(fAttr);
            }
        }

    }

    static final long serialVersionUID = 1297326755;

    private static final Logger logger = Logger.getLogger(AdminHandler.class.getName());

    private File confFile;

    private long configTS;

    private File adminFile;

    private long adminTS;

    private HashSet<String> dnTable;

    private AdminConfigParser internalParser;

    public void init() {
        super.init();

        internalParser = new AdminConfigParser();

        dnTable = new HashSet<String>();

        configTS = 0;

        adminTS = 0;

        try {
            Hashtable<String, String> env = new Hashtable<String, String>(0);
            env.put(Context.INITIAL_CONTEXT_FACTORY, "org.apache.naming.java.javaURLContextFactory");
            env.put(Context.URL_PKG_PREFIXES, "org.apache.naming");

            Context context = new InitialContext(env);
            context = (Context) context.lookup("java:comp/env");

            String tmpFilename = (String) context.lookup("configuration_provider_url");
            if (tmpFilename == null) {
                logger.error("Missing configuration_provider_url parameter");
                return;
            }
            confFile = new File(tmpFilename);

        } catch (Exception ex) {
            logger.error(ex.getMessage(), ex);
            confFile = null;
            adminFile = null;
        }

    }

    public void invoke(MessageContext msgContext)
        throws AxisFault {

        if (confFile == null) {
            throw (AxisFault)this.getAuthorizationFault("Cannot verify user's permissions");
        }

        try {
            checkUpdate();

            if (adminFile == null) {
                return;
            }

        } catch (Exception ex) {
            logger.error(ex.getMessage(), ex);
            throw (AxisFault)this.getAuthorizationFault("Cannot verify user's permissions");
        }

        InitSecurityContext.init();
        SecurityInfo secInfo = SecurityInfoContainer.getSecurityInfo();
        String dnX500 = secInfo.getClientX500Name();
        Boolean isAdmin = new Boolean(dnTable.contains(dnX500));
        logger.debug("Admin test for " + dnX500 + ": " + isAdmin);

        msgContext.setProperty(AuthorizationHandler.IS_ADMIN, isAdmin);

    }

    private synchronized void checkUpdate()
        throws IOException, SAXException, ParserConfigurationException {
        if (confFile.lastModified() > configTS) {
            adminFile = null;
            parseConfigFile();
            configTS = confFile.lastModified();
        }

        if (adminFile != null && adminFile.lastModified() > adminTS) {
            fillinDNTable();
            adminTS = adminFile.lastModified();
        }

    }

    private void fillinDNTable()
        throws IOException {

        dnTable.clear();

        BufferedReader reader = null;
        try {
            reader = new BufferedReader(new FileReader(adminFile));
            String line = reader.readLine();
            while (line != null) {
                line = line.trim();

                if (line.length() > 0 && !line.startsWith("#")) {
                    if (line.startsWith("\"") && line.endsWith("\""))
                        line = line.substring(1, line.length() - 1);
                    dnTable.add(line);
                    logger.debug("Registered DN: " + line);
                }
                line = reader.readLine();
            }
        } finally {
            if (reader != null) {
                try {
                    reader.close();
                } catch (Exception ex) {
                }
            }
        }
    }

    private void parseConfigFile()
        throws IOException, SAXException, ParserConfigurationException {

        FileReader xmlReader = null;

        try {
            SAXParserFactory spf = SAXParserFactory.newInstance();
            spf.setValidating(false);
            SAXParser sp = spf.newSAXParser();

            xmlReader = new FileReader(confFile);
            InputSource input = new InputSource(xmlReader);
            input.setSystemId("file://" + confFile);

            logger.debug("Parsing file " + confFile);

            sp.parse(input, this.internalParser);
        } finally {
            if (xmlReader != null)
                try {
                    xmlReader.close();
                } catch (Exception ex) {
                }
        }
    }

    /*
     * TODO temporary workaround
     *      Object as return type instead of AxisFault
     *      investigate VerifyError with axis 1.4
     */ 
    protected Object getAuthorizationFault(String message) {
        AuthorizationFault fault = new AuthorizationFault();
        fault.setMethodName("invoke");
        fault.setDescription(message);
        fault.setErrorCode("0");
        fault.setFaultCause(message);
        fault.setTimestamp(Calendar.getInstance());
        return fault;
    }

}
