/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

/*
 * 
 * Authors: Paolo Andreetto, <paolo.andreetto@pd.infn.it>
 *
 */

package org.glite.ce.commonj.authz;

import java.io.File;
import java.lang.reflect.Constructor;
import java.util.ArrayList;

import org.apache.log4j.Logger;
import org.glite.ce.commonj.configuration.basic.ConfigItem;
import org.glite.ce.commonj.configuration.basic.ConfigParser;
import org.glite.ce.commonj.configuration.basic.TriggerManager;
import org.xml.sax.Attributes;
import org.xml.sax.SAXParseException;

public class AuthZConfigParser implements ConfigParser{
    
    private static Logger logger = Logger.getLogger(AuthZConfigParser.class.getName());

    protected String chainName;
    protected ArrayList<ServiceInterceptor> plugins;
    protected ServiceInterceptor currentPlugin;
    protected TriggerManager triggerManager;
    protected ClassLoader loader;

    public AuthZConfigParser(TriggerManager tMan){
        triggerManager = tMan;
        loader = this.getClass().getClassLoader();        
    }

    public String getObjectName(String uri, String name,
            String qName, Attributes attributes) throws SAXParseException {
        if (qName.equals("authzchain")) {
            String result = attributes.getValue("name");
            if( result==null )
                throw new SAXParseException("Missing authZ chain name", null);

            return result;
        }
        return null;
    }

    public void startElement(String uri, String name, String qName,
            Attributes attributes, ConfigItem prevObj) throws SAXParseException {

        if( qName.equals("authzchain") ){

            chainName = attributes.getValue("name");
            plugins = new ArrayList<ServiceInterceptor>();
        
        }else if (qName.equals("plugin")){
            
            try{
                Class<?> interceptorClass = Class.forName(attributes.getValue("classname"));
                Class<?>[] constrArgClass = new Class<?>[]{String.class};
                Constructor<?> constr = interceptorClass.getConstructor(constrArgClass);
                Object[] constrArgValue = new Object[]{attributes.getValue("name")};
                currentPlugin = (ServiceInterceptor) constr.newInstance(constrArgValue);

            }catch(Exception ex){
                logger.error(ex.getMessage(), ex);
                throw new SAXParseException("Cannot load interceptor chain: " 
                                                                + ex.getMessage(), null);
            }
            
        }else if(qName.equals("parameter")){
            
            String pName = attributes.getValue("name");
            if( pName==null)
                throw new SAXParseException("Missing name attribute for parameter", null);
            
            String pValue = attributes.getValue("value");
            if( pValue==null)
                throw new SAXParseException("Missing value for parameter " + pName, null);

            try{
                currentPlugin.setProperty(pName, pValue);                
            }catch(InitializeException initEx){
                throw new SAXParseException(initEx.getMessage(), null);
            }
            
        }

    }

    public void endElement(String uri, String name, String qName, 
            ConfigItem currentObj) throws SAXParseException {
        
        if( qName.equals("authzchain") ){
            
            Long tmpl = new Long(System.currentTimeMillis());
            boolean changed = false;
            
            ServiceAuthorizationChain oldChain = (ServiceAuthorizationChain)currentObj.getContent();
            ServiceAuthorizationChain newChain = new ServiceAuthorizationChain();
            newChain.initialize(chainName, plugins);

            if( oldChain==null ){
                currentObj.put("creationtime", tmpl);
                currentObj.put("category", "authzchain");
                logger.debug("Register new service chain " + chainName);
                changed = true;
            }else if( !newChain.equals(oldChain) ){
                logger.debug("Detected changes in service chain: " + oldChain.getId());
                changed = true;

                File[] oldTriggers = oldChain.getAllTriggers();
                for(int k=0; k<oldTriggers.length; k++){
                    triggerManager.removeTrigger(oldTriggers[k]);
                    logger.debug("Removed trigger " + oldTriggers[k].toString());
                }
                
            }            

            if( changed ){
                currentObj.setContent(newChain);
                currentObj.put("modificationtime", tmpl);
                
                File[] newTriggers = newChain.getAllTriggers();
                for(int k=0; k<newTriggers.length; k++){
                    triggerManager.addTrigger(newTriggers[k]);
                    logger.debug("Added trigger " + newTriggers[k].toString());
                }
            }
            
        }else if (qName.equals("plugin")) {
            logger.debug("Inserted new plugin: " + currentPlugin.getId());
            plugins.add(currentPlugin);
        }

    }

    public void characters(char[] chars, int start, int length, ConfigItem currentContext) 
        throws SAXParseException{
        
    }
}


