/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

/*
 * 
 * Authors: Paolo Andreetto, <paolo.andreetto@pd.infn.it>
 *
 */

package org.glite.ce.commonj.authz;

import java.security.cert.X509Certificate;
import java.util.Calendar;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;
import javax.naming.directory.BasicAttributes;
import javax.naming.directory.DirContext;
import javax.naming.directory.InitialDirContext;
import javax.naming.directory.SearchResult;
import javax.security.auth.Subject;
import javax.servlet.http.HttpServletRequest;
import javax.xml.namespace.QName;

import org.apache.axis.AxisFault;
import org.apache.axis.MessageContext;
import org.apache.axis.description.OperationDesc;
import org.apache.axis.handlers.BasicHandler;
import org.apache.axis.transport.http.HTTPConstants;
import org.apache.log4j.Logger;
import org.glite.ce.faults.AuthorizationFault;
import org.glite.security.SecurityContext;
import org.glite.security.SecurityInfoContainer;
import org.glite.security.util.axis.InitSecurityContext;
import org.glite.voms.FQAN;
import org.glite.voms.VOMSAttribute;
import org.glite.voms.VOMSValidator;
import org.glite.voms.ac.ACValidator;

public class AuthorizationHandler
    extends BasicHandler {
    private static final Logger logger = Logger.getLogger(AuthorizationHandler.class.getName());

    private static final long serialVersionUID = 1202907233;

    private static final String CONFIG_REQ_LABEL = "config_required";

    private static final String HANDLER_TYPE_LABEL = "type";

    public static final String USER_VOMSATTRS_LABEL = "user.vomsattrs";

    public static final String USERDN_X500_LABEL = "user.dn.X500";

    public static final String USERDN_RFC2253_LABEL = "user.dn.RFC2253";

    public static final String USER_CERTCHAIN_LABEL = "user.certificate.chain";

    public static final String IS_ADMIN = "IS_ADMIN";

    public static final String REMOTE_REQUEST_ADDRESS = "REMOTE_REQUEST_ADDRESS";

    public static final String LOCAL_USER_ID = "local.user.id";

    public static final String LOCAL_GROUP_ID = "local.group.id";

    public static final String LOCAL_GROUPS_LIST = "local.groups.list";

    public static final String JOB_ADMIN = "job.admin.role";

    public DirContext rootCtx;

    public String handlerType;

    public boolean confReq;

    public AuthorizationHandler() {
        super();
    }

    public void init() {
        super.init();

        handlerType = (String) this.getOption(HANDLER_TYPE_LABEL);

        String tmpCR = (String) this.getOption(CONFIG_REQ_LABEL);
        confReq = (tmpCR != null && tmpCR.equalsIgnoreCase("true"));

        rootCtx = null;
        Hashtable<String, String> env = new Hashtable<String, String>(0);
        env.put(Context.INITIAL_CONTEXT_FACTORY, "org.apache.naming.java.javaURLContextFactory");
        env.put(Context.URL_PKG_PREFIXES, "org.apache.naming");

        try {
            Context context = new InitialContext(env);
            context = (Context) context.lookup("java:comp/env");

            String confProvider = (String) context.lookup("configuration_provider");
            String confProviderURL = (String) context.lookup("configuration_provider_url");

            if (confProvider == null) {
                logger.error("Missing configuration_provider parameter");
                return;
            }

            if (confProviderURL == null) {
                logger.error("Missing configuration_provider_url parameter");
                return;
            }

            env.put(Context.INITIAL_CONTEXT_FACTORY, confProvider);
            env.put(Context.PROVIDER_URL, confProviderURL);
            rootCtx = (DirContext) new InitialDirContext(env);

        } catch (NamingException nEx) {
            logger.error(nEx.getMessage(), nEx);
        }
    }

    protected void check(javax.xml.rpc.handler.MessageContext msgContext)
        throws AuthorizationException {

        if (rootCtx == null || handlerType == null) {
            throw new AuthorizationException("Authorization misconfigured");
        }

        long currentTS = System.currentTimeMillis();
        SecurityContext secInfo = (SecurityContext) SecurityInfoContainer.getSecurityInfo();

        String dnRFC2253 = secInfo.getClientDN().getRFCDNv2();
        String dnX500 = secInfo.getClientX500Name();

        if (dnRFC2253 == null || dnRFC2253.equals("") || dnX500 == null || dnX500.equals("")) {
            throw new AuthorizationException("Cannot retrieve credentials");
        }

        Subject subject = new Subject();
        subject.getPrincipals().add(secInfo.getClientX500Principal());
        subject.getPrincipals().add(new javax.security.auth.x500.X500Principal(dnRFC2253));

        msgContext.setProperty(USERDN_X500_LABEL, dnX500);
        msgContext.setProperty(USERDN_RFC2253_LABEL, dnRFC2253);

        X509Certificate[] userCertChain = secInfo.getClientCertChain();
        msgContext.setProperty(USER_CERTCHAIN_LABEL, userCertChain);
        subject.getPublicCredentials().add(userCertChain);

        ACValidator acValidator = AuthZContextListener.getACValidator();
        VOMSValidator mainValidator = new VOMSValidator(userCertChain, acValidator);
        mainValidator.validate();

        List<VOMSAttribute> vomsList = (List<VOMSAttribute>) mainValidator.getVOMSAttributes();
        msgContext.setProperty(USER_VOMSATTRS_LABEL, vomsList);

        QName operation = this.getOperation(msgContext);
        if (operation == null) {
            throw new AuthorizationException("Cannot recognize operation");
        }
        logger.debug("Checking operation " + operation);
        ServiceAuthorizationInterface authzBox = null;

        try {

            BasicAttributes queryAttrs = new BasicAttributes("category", handlerType);
            NamingEnumeration<SearchResult> queryResult = rootCtx.search("", queryAttrs);
            if (queryResult.hasMoreElements()) {

                Object tmpo = queryResult.next().getObject();
                if (tmpo instanceof ServiceAuthorizationInterface) {

                    authzBox = (ServiceAuthorizationInterface) tmpo;

                } else if (tmpo instanceof ServiceAuthorizationFactory) {

                    ServiceAuthorizationFactory authzFactory = (ServiceAuthorizationFactory) tmpo;
                    authzBox = authzFactory.getInstance();

                }
            }

        } catch (Exception ex) {
            logger.error(ex.getMessage(), ex);
        }

        if (authzBox == null) {
            if (confReq) {
                throw new AuthorizationException("Authorization layer is not configured");
            } else {
                logger.debug("Cannot retrieve authorization box for " + handlerType);
                return;
            }
        }

        StringBuffer reqLogInfo = new StringBuffer("request for operation=");
        reqLogInfo.append(operation);
        reqLogInfo.append("; ");

        HttpServletRequest httpReq = (HttpServletRequest) msgContext.getProperty(HTTPConstants.MC_HTTP_SERVLETREQUEST);
        if (httpReq != null) {
            reqLogInfo.append("REMOTE_REQUEST_ADDRESS=");
            reqLogInfo.append(httpReq.getRemoteAddr());
            reqLogInfo.append("; ");

            msgContext.setProperty(REMOTE_REQUEST_ADDRESS, httpReq.getRemoteAddr());
        }

        reqLogInfo.append("USER_DN=").append(dnRFC2253).append("; ");

        if (vomsList != null && vomsList.size() > 0) {
            reqLogInfo.append("USER_FQAN={ ");

            for (VOMSAttribute vomsAttr : vomsList) {
                List<FQAN> list = vomsAttr.getListOfFQAN();
                if (list != null) {
                    for (FQAN fqan : list) {
                        reqLogInfo.append(fqan).append("; ");
                    }
                }
            }

            reqLogInfo.append("}; ");

            msgContext.setProperty(REMOTE_REQUEST_ADDRESS, httpReq.getRemoteAddr());
        }

        if (authzBox.isPermitted(subject, msgContext, operation)) {
            reqLogInfo.append(" AUTHORIZED!");
            logger.info(reqLogInfo);
            logger.debug("PROFILING " + handlerType + " " + (System.currentTimeMillis() - currentTS));
            return;
        }

        reqLogInfo.append(" NOT AUTHORIZED!");
        logger.info(reqLogInfo);
        logger.debug("PROFILING " + handlerType + " " + (System.currentTimeMillis() - currentTS));

        throw new AuthorizationException("User " + dnRFC2253 + " not authorized for operation " + operation);
    }

    protected QName getOperation(javax.xml.rpc.handler.MessageContext context) {

        MessageContext msgContext = (MessageContext) context;
        OperationDesc operation = msgContext.getOperation();
        javax.xml.soap.SOAPBody msgBody = null;

        if (operation == null) {
            /*
             * This workaround may be related to the definition of the handler
             * in the global section of the wsdd
             */
            try {
                msgBody = msgContext.getMessage().getSOAPBody();
            } catch (Exception ex) {
                logger.warn(ex.getMessage());
            }
            operation = msgContext.getOperation();
        }

        if (operation == null) {
            if (logger.isDebugEnabled() && msgBody != null) {
                logger.debug(msgBody);
            }
            return null;
        }
        return operation.getElementQName();
    }

    /*
     * TODO temporary workaround
     *      Object as return type instead of AxisFault
     *      investigate VerifyError with axis 1.4
     */
    protected Object getAuthorizationFault(String message) {
        AuthorizationFault fault = new AuthorizationFault();
        fault.setMethodName("invoke");
        fault.setDescription(message);
        fault.setErrorCode("0");
        fault.setFaultCause(message);
        fault.setTimestamp(Calendar.getInstance());
        return fault;
    }

    public void invoke(MessageContext msgContext)
        throws AxisFault {
        InitSecurityContext.init();
        try {
            this.check(msgContext);
        } catch (AuthorizationException authEx) {
            logger.error(authEx.getMessage());
            throw (AxisFault)this.getAuthorizationFault(authEx.getMessage());
        }
    }
}
