/*
 * Copyright (c) Members of the EGEE Collaboration. 2004. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 *
 *     http://www.apache.org/licenses/LICENSE-2.0 
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 */

/*
 * 
 * Authors: Paolo Andreetto, <paolo.andreetto@pd.infn.it>
 *
 */

package org.glite.ce.commonj.authz.argus;

import java.io.File;

import org.apache.log4j.Logger;
import org.glite.authz.pep.client.config.PEPClientConfigurationException;
import org.glite.ce.commonj.configuration.basic.ConfigItem;
import org.glite.ce.commonj.configuration.basic.ConfigParser;
import org.glite.ce.commonj.configuration.basic.TriggerManager;
import org.xml.sax.Attributes;
import org.xml.sax.SAXParseException;

public class ArgusPEPConfigParser
    implements ConfigParser {

    private static Logger logger = Logger.getLogger(ArgusPEPConfigParser.class.getName());

    private static String ARGUS_MAIN_TAG = "argus-pep";

    private static String PAP_ENDPOINT_TAG = "pap-endpoint";

    private static String PAP_URL_ATTR = "url";

    private static String CERT_FILE_ATTR = "cert";

    private static String KEY_FILE_ATTR = "key";

    private static String PWD_ATTR = "passwd";

    private static String TIMEOUT_ATTR = "timeout";

    private static String CONN_HOST_ATTR = "connection_per_host";

    private static String MAX_CONN_ATTR = "max_connection";

    private static String CA_DIR_ATTR = "ca_dir";

    private static String RES_ID_ATTR = "resource_id";

    private static String MAPPING_ATTR = "mapping_class";

    public final static String ARGUS_CATEGORY = "argus-pep";

    private PEPConfigurationItem pepConfig;

    public ArgusPEPConfigParser(TriggerManager tMan) {
    }

    public String getObjectName(String uri, String name, String qName, Attributes attributes)
        throws SAXParseException {
        if (qName.equals(ARGUS_MAIN_TAG)) {
            String result = attributes.getValue("name");
            if (result == null)
                throw new SAXParseException("Missing argus-pep name", null);

            return result;
        }
        return null;
    }

    public void startElement(String uri, String name, String qName, Attributes attributes, ConfigItem prevObj)
        throws SAXParseException {

        if (qName.equals(ARGUS_MAIN_TAG)) {

            logger.debug("Parsing argus pep configuration");

            pepConfig = new PEPConfigurationItem();

            String certFilename = (String) attributes.getValue(CERT_FILE_ATTR);
            if (certFilename == null || !(new File(certFilename).exists())) {
                throw new SAXParseException("Missing or wrong certificate filename", null);
            }

            String keyFilename = (String) attributes.getValue(KEY_FILE_ATTR);
            if (keyFilename == null || !(new File(keyFilename).exists())) {
                throw new SAXParseException("Missing or wrong key filename", null);
            }

            String pwd = (String) attributes.getValue(PWD_ATTR);
            if (pwd == null) {
                pwd = "";
            }

            String timeoutStr = (String) attributes.getValue(TIMEOUT_ATTR);
            if (timeoutStr != null) {
                try {
                    pepConfig.setConnectionTimeout(Integer.parseInt(timeoutStr));
                } catch (Exception ex) {
                    throw new SAXParseException("Cannot parse " + TIMEOUT_ATTR, null);
                }
            }

            String connHostStr = (String) attributes.getValue(CONN_HOST_ATTR);
            if (connHostStr != null) {
                try {
                    pepConfig.setMaxConnectionsPerHost(Integer.parseInt(connHostStr));
                } catch (Exception ex) {
                    throw new SAXParseException("Cannot parse " + CONN_HOST_ATTR, null);
                }
            }

            String maxConnStr = (String) attributes.getValue(MAX_CONN_ATTR);
            if (maxConnStr != null) {
                try {
                    pepConfig.setMaxTotalConnections(Integer.parseInt(maxConnStr));
                } catch (Exception ex) {
                    throw new SAXParseException("Cannot parse " + MAX_CONN_ATTR, null);
                }
            }

            try {
                pepConfig.setKeyMaterial(certFilename, keyFilename, pwd);
            } catch (PEPClientConfigurationException pEx) {
                logger.error(pEx.getMessage(), pEx);
                throw new SAXParseException(pEx.getMessage(), null);
            }

            String caDir = (String) attributes.getValue(CA_DIR_ATTR);
            if (caDir == null) {
                caDir = "/etc/grid-security/certificates";
            }
            if (!(new File(caDir)).isDirectory()) {
                throw new SAXParseException("Missing or wrong CA directory", null);
            }
            try {
                pepConfig.setTrustMaterial(caDir);
            } catch (Exception ex) {
                logger.error(ex.getMessage(), ex);
                throw new SAXParseException(ex.getMessage(), null);
            }

            String resID = (String) attributes.getValue(RES_ID_ATTR);
            if (resID != null) {
                pepConfig.setResourceID(resID);
            } else {
                throw new SAXParseException("Missing attribute " + RES_ID_ATTR, null);
            }

            String mapClass = (String) attributes.getValue(MAPPING_ATTR);
            if (mapClass == null)
                throw new SAXParseException("Missing mapping class", null);
            pepConfig.setMappingClass(mapClass);

        } else if (qName.equals(PAP_ENDPOINT_TAG)) {
            String papURL = (String) attributes.getValue(PAP_URL_ATTR);
            logger.debug("Registered pap endpoint " + papURL);
            pepConfig.addPEPDaemonEndpoint(papURL);
        }
    }

    public void endElement(String uri, String name, String qName, ConfigItem currentObj)
        throws SAXParseException {

        if (qName.equals(ARGUS_MAIN_TAG)) {
            Long tmpl = new Long(System.currentTimeMillis());
            boolean changed = false;

            PEPConfigurationItem oldConfig = (PEPConfigurationItem) currentObj.getContent();

            if (oldConfig == null) {
                currentObj.put("creationtime", tmpl);
                currentObj.put("category", ARGUS_CATEGORY);
                logger.debug("Register new Argus pep client");
                changed = true;
            } else if (!pepConfig.equals(oldConfig)) {
                logger.debug("Detected changes in Argus pep client");
                changed = true;
            }

            if (changed) {
                currentObj.setContent(pepConfig);
                currentObj.put("modificationtime", tmpl);
            }
        }

    }

    public void characters(char[] chars, int start, int length, ConfigItem currentContext)
        throws SAXParseException {
    }

}
